# more details: https://github.com/tmbdev-archive/webdataset-examples
import sys
sys.path.append("..")
import argparse
import os
import random
import shutil
import time
import warnings
import logging

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.optim as optim
import torch.multiprocessing as mp
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models
import webdataset as wds

from util import SynchronousValidation

from optim.adadqh import AdaDQH
from optim.optim_adahessian import Adahessian
from optim.adabelief import AdaBelief

LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
logging.basicConfig(filename='./output.log', level=logging.INFO, format=LOG_FORMAT)
logging.captureWarnings(True)

model_names = sorted(name for name in models.__dict__
    if name.islower() and not name.startswith("__")
    and callable(models.__dict__[name]))

parser = argparse.ArgumentParser(description='PyTorch ImageNet Training')
parser.add_argument('--data', metavar='DIR', default='./data',
                    help='path to dataset')
parser.add_argument('--loader', default='wds', help='loader to use: orig, wds')
parser.add_argument('--shuffle', type=int, default=1000, help='shuffle buffer size for WebDataset')
parser.add_argument('--trainshards', default='/penrose_input_data/imagenet/shards/imagenet-train-{000000..000146}.tar', help='path/URL for ImageNet shards',
)
parser.add_argument('--valshards', default='/penrose_input_data/imagenet/val_shards/imagenet-val-{000000..000022}.tar', help='path/URL for ImageNet shards',
)
parser.add_argument('--trainsize', type=int, default=1281167, help='ImageNet training set size')
parser.add_argument('--valsize', type=int, default=50000, help='ImageNet validating set size')
parser.add_argument('--augmentation', default='full')
parser.add_argument('-a', '--arch', metavar='ARCH', default='resnet18',
                    choices=model_names,
                    help='model architecture: ' +
                        ' | '.join(model_names) +
                        ' (default: resnet18)')
parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                    help='number of data loading workers (default: 4)')
parser.add_argument('--epochs', default=90, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N',
                    help='mini-batch size (default: 256), this is the total '
                         'batch size of all GPUs on the current node when '
                         'using Data Parallel or Distributed Data Parallel')
parser.add_argument('--lr', '--learning-rate', default=0.1, type=float,
                    metavar='LR', help='initial learning rate', dest='lr')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--wd', '--weight-decay', default=1e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)',
                    dest='weight_decay')
parser.add_argument('--weight-decouple', type=bool, default=True,
                    help='weight decouple (default: True)')
parser.add_argument('-p', '--print-freq', default=10, type=int,
                    metavar='N', help='print frequency (default: 10)')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
parser.add_argument('--pretrained', dest='pretrained', action='store_true',
                    help='use pre-trained model')
parser.add_argument('--world-size', default=-1, type=int,
                    help='number of nodes for distributed training')
parser.add_argument('--rank', default=-1, type=int,
                    help='node rank for distributed training')
parser.add_argument('--dist-url', default='tcp://224.66.41.62:23456', type=str,
                    help='url used to set up distributed training')
parser.add_argument('--dist-backend', default='nccl', type=str,
                    help='distributed backend')
parser.add_argument('--seed', default=None, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--gpu', default=None, type=int,
                    help='GPU id to use.')
parser.add_argument('--multiprocessing-distributed', action='store_true',
                    help='Use multi-processing distributed training to launch '
                         'N processes per node, which has N GPUs. This is the '
                         'fastest way to use PyTorch for either single node or '
                         'multi node data parallel training')

parser.add_argument('--optimizer', type=str, default='adadqh',
                    help='choose optim')
parser.add_argument('--eps', type=float, default=1e-8,
                    help='choose epsilon')

best_acc1 = 0


def main():
    args = parser.parse_args()
    for arg in vars(args):
        logging.info(f"{arg} {getattr(args, arg)}")

    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        cudnn.deterministic = True
        warnings.warn('You have chosen to seed training. '
                      'This will turn on the CUDNN deterministic setting, '
                      'which can slow down your training considerably! '
                      'You may see unexpected behavior when restarting '
                      'from checkpoints.')

    if args.gpu is not None:
        warnings.warn('You have chosen a specific GPU. This will completely '
                      'disable data parallelism.')

    if args.dist_url == "env://" and args.world_size == -1:
        args.world_size = int(os.environ["WORLD_SIZE"])

    args.distributed = args.world_size > 1 or args.multiprocessing_distributed

    ngpus_per_node = torch.cuda.device_count()
    manager = mp.Manager()
    validator = SynchronousValidation(worker_num=ngpus_per_node, manager=manager)

    if args.multiprocessing_distributed:
        # Since we have ngpus_per_node processes per node, the total world_size
        # needs to be adjusted accordingly
        args.world_size = ngpus_per_node * args.world_size
        # Use torch.multiprocessing.spawn to launch distributed processes: the
        # main_worker process function
        mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, args, validator))
    else:
        # Simply call main_worker function
        main_worker(args.gpu, ngpus_per_node, args)


def main_worker(gpu, ngpus_per_node, args, validator):
    global best_acc1
    args.gpu = gpu

    if args.gpu is not None:
        logging.info("Use GPU: {} for training".format(args.gpu))

    if args.distributed:
        if args.dist_url == "env://" and args.rank == -1:
            args.rank = int(os.environ["RANK"])
        if args.multiprocessing_distributed:
            # For multiprocessing distributed training, rank needs to be the
            # global rank among all the processes
            args.rank = args.rank * ngpus_per_node + gpu
        dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                world_size=args.world_size, rank=args.rank)
    # create model
    if args.pretrained:
        logging.info("=> using pre-trained model '{}'".format(args.arch))
        model = models.__dict__[args.arch](pretrained=True)
    else:
        logging.info("=> creating model '{}'".format(args.arch))
        model = models.__dict__[args.arch]()

    if not torch.cuda.is_available():
        logging.info('using CPU, this will be slow')
    elif args.distributed:
        # For multiprocessing distributed, DistributedDataParallel constructor
        # should always set the single device scope, otherwise,
        # DistributedDataParallel will use all available devices.
        if args.gpu is not None:
            torch.cuda.set_device(args.gpu)
            model.cuda(args.gpu)
            # When using a single GPU per process and per
            # DistributedDataParallel, we need to divide the batch size
            # ourselves based on the total number of GPUs we have
            args.batch_size = int(args.batch_size / ngpus_per_node)
            # args.workers = int((args.workers + ngpus_per_node - 1) / ngpus_per_node)
            logging.info("batch_size: {}".format(args.batch_size))
            logging.info("workers: {}".format(args.workers))
            model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
        else:
            model.cuda()
            # DistributedDataParallel will divide and allocate batch_size to all
            # available GPUs if device_ids are not set
            model = torch.nn.parallel.DistributedDataParallel(model)
    elif args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        model = model.cuda(args.gpu)
    else:
        # DataParallel will divide and allocate batch_size to all available GPUs
        if args.arch.startswith('alexnet') or args.arch.startswith('vgg'):
            model.features = torch.nn.DataParallel(model.features)
            model.cuda()
        else:
            model = torch.nn.DataParallel(model).cuda()

    logging.info('    Total params: %.2fM' % (sum(p.numel() for p in model.parameters()) / 1000000.0))

    # define loss function (criterion) and optimizer
    criterion = nn.CrossEntropyLoss().cuda(args.gpu)

    if args.optimizer == 'sgd':
        optimizer = optim.SGD(
            model.parameters(),
            lr=args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay)
    elif args.optimizer == 'adam':
        optimizer = optim.Adam(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay)
    elif args.optimizer == 'adamw':
        logging.info(
            'For AdamW, we automatically correct the weight decay term for you! If this is not what you want, '
            'please modify the code!')
        args.weight_decay = args.weight_decay / args.lr
        optimizer = optim.AdamW(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay)
    elif args.optimizer == 'adahessian':
        logging.info(
            'For AdaHessian, we use the decoupled weight decay as AdamW. Here we automatically correct this for you! '
            'If this is not what you want, please modify the code!')
        args.weight_decay = args.weight_decay / args.lr
        optimizer = Adahessian(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay,
            single_gpu=False)
    elif args.optimizer == 'adabelief':
        optimizer = AdaBelief(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay,
            weight_decouple=args.weight_decouple,
        )
    elif args.optimizer == 'adadqh':
        optimizer = AdaDQH(
            model.parameters(),
            lr=args.lr,
            eps=args.eps,
            weight_decay=args.weight_decay,
            weight_decouple=args.weight_decouple,
        )
    else:
        raise Exception('We do not support this optimizer yet!!')

    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            logging.info("=> loading checkpoint '{}'".format(args.resume))
            if args.gpu is None:
                checkpoint = torch.load(args.resume)
            else:
                # Map model to be loaded to specified single gpu.
                loc = 'cuda:{}'.format(args.gpu)
                checkpoint = torch.load(args.resume, map_location=loc)
            args.start_epoch = checkpoint['epoch']
            best_acc1 = checkpoint['best_acc1']
            if args.gpu is not None:
                # best_acc1 may be from a checkpoint from a different GPU
                best_acc1 = best_acc1.to(args.gpu)
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            logging.info("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.resume, checkpoint['epoch']))
        else:
            logging.info("=> no checkpoint found at '{}'".format(args.resume))

    cudnn.benchmark = True

    train_loader = eval(f"make_train_loader_{args.loader}")(args)

    # val_loader = make_val_loader(args)
    val_loader = eval(f"make_val_loader_{args.loader}")(args)

    if args.evaluate:
        validate(gpu, val_loader, model, criterion, args, validator)
        return

    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            sampler = getattr(train_loader, "sampler", None)
            if sampler is not None and hasattr(sampler, "set_epoch"):
                sampler.set_epoch(epoch)
        adjust_learning_rate(optimizer, epoch, args)

        # train for one epoch
        train(gpu, train_loader, model, criterion, optimizer, epoch, args)

        # evaluate on validation set
        acc1 = validate(gpu, val_loader, model, criterion, args, validator)
        logging.info(f'pid: {gpu} acc1 of Epoch {epoch + 1}: {acc1}')
        if gpu == 0:
            acc1, acc5 = validator.cal_accu_metrics()
            logging.info(f'total acc1 in pid: {gpu} of Epoch {epoch + 1}: {acc1}')
            validator.clear()

            # remember best acc@1 and save checkpoint
            is_best = acc1 > best_acc1
            best_acc1 = max(acc1, best_acc1)

            if not args.multiprocessing_distributed or (args.multiprocessing_distributed
                    and args.rank % ngpus_per_node == 0):
                save_checkpoint({
                    'epoch': epoch + 1,
                    'arch': args.arch,
                    'state_dict': model.state_dict(),
                    'best_acc1': best_acc1,
                    'optimizer' : optimizer.state_dict(),
                }, is_best)
        validator.done_evaluation(gpu)
    if gpu == 0:
        logging.info(f'Best Acc: {best_acc1}')


def make_train_transform(args):
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

    if args.augmentation == "full":
        return transforms.Compose(
            [
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                normalize,
            ]
        )
    elif args.augmentation == "simple":
        logging.info("=> using simple augmentation")
        return transforms.Compose(
            [transforms.Resize(256), transforms.CenterCrop(224), transforms.ToTensor(), normalize,]
        )


def make_val_transform(args):
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

    return transforms.Compose(
        [transforms.Resize(256), transforms.CenterCrop(224), transforms.ToTensor(), normalize,]
    )


def make_train_loader_orig(args):
    logging.info("=> using file based loader")
    traindir = os.path.join(args.data, "train")
    train_transform = make_train_transform(args)
    train_dataset = datasets.ImageFolder(traindir, train_transform)
    args.trainsize = len(train_dataset)

    if args.distributed:
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
    else:
        train_sampler = None

    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        shuffle=(train_sampler is None),
        num_workers=args.workers,
        pin_memory=True,
        sampler=train_sampler,
    )

    return train_loader


def identity(x):
    return x


def worker_urls(urls):
    result = wds.worker_urls(urls)
    print("worker_urls returning", len(result), "of", len(urls), "urls", file=sys.stderr)
    return result


def make_train_loader_wds(args):
    logging.info("=> using WebDataset loader in train")
    train_transform = make_train_transform(args)
    num_batches = args.trainsize // args.batch_size
    train_dataset = (
        wds.WebDataset(args.trainshards)
        .shuffle(args.shuffle)
        .decode("pil")
        .to_tuple("jpg;png;jpeg cls")
        .map_tuple(train_transform, identity)
    )
    if args.distributed:
        # It's good to avoid partial batches when using DistributedDataParallel.
        train_dataset = train_dataset.batched(args.batch_size, partial=False)
    else:
        train_dataset = train_dataset.batched(args.batch_size)
    # WebLoader is just the regular DataLoader with the same convenience methods
    # that WebDataset has.
    train_loader = wds.WebLoader(
        train_dataset, batch_size=None, shuffle=False, num_workers=args.workers,
    )
    # train_loader = torch.utils.data.DataLoader(
    #     train_dataset, batch_size=None, shuffle=False, num_workers=args.workers,
    # )
    if args.distributed:
        # With DDP, we need to make sure that all nodes get the same number of batches
        logging.info("# batches per node = {}".format(num_batches // args.world_size))
        train_loader = train_loader.ddp_equalize(num_batches)
        # This only sets the value returned by the len() function; nothing else uses it,
        # but some frameworks care about it.
        logging.info(f"world size: {args.world_size}  batch size: {args.batch_size}")
        train_loader.length = num_batches //args.world_size
    else:
        logging.info(f"# num_batches = {num_batches}")
        train_loader.length = num_batches
    return train_loader


def make_val_loader_orig(args):
    valdir = os.path.join(args.data, "val")
    val_transform = make_val_transform(args)
    val_dataset = datasets.ImageFolder(valdir, val_transform)
    val_loader = torch.utils.data.DataLoader(
        val_dataset, batch_size=args.batch_size, shuffle=False, num_workers=args.workers, pin_memory=True
    )
    return val_loader


def make_val_loader_wds(args):
    logging.info("=> using WebDataset loader in val")
    val_transform = make_val_transform(args)
    num_batches = args.valsize // args.batch_size + 1
    val_dataset = (
        wds.WebDataset(args.valshards)
        .decode("pil")
        .to_tuple("jpg;png;jpeg cls")
        .map_tuple(val_transform, identity)
    )
    if args.distributed:
        # It's good to avoid partial batches when using DistributedDataParallel.
        val_dataset = val_dataset.batched(args.batch_size, partial=True)
    else:
        val_dataset = val_dataset.batched(args.batch_size)
    # WebLoader is just the regular DataLoader with the same convenience methods
    # that WebDataset has.
    val_loader = wds.WebLoader(
        val_dataset, batch_size=None, shuffle=False, num_workers=args.workers,
    )
    # val_loader = torch.utils.data.DataLoader(
    #     train_dataset, batch_size=None, shuffle=False, num_workers=args.workers,
    # )
    if args.distributed:
        val_loader.length = num_batches // args.world_size
    else:
        val_loader.length = num_batches

    return val_loader


def train(pid, train_loader, model, criterion, optimizer, epoch, args):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        train_loader.length,
        [batch_time, data_time, losses, top1, top5],
        prefix="pid: {} Epoch: [{}]".format(pid, epoch))

    # switch to train mode
    model.train()

    end = time.time()
    for i, (images, target) in enumerate(train_loader):
        # measure data loading time
        data_time.update(time.time() - end)

        if args.gpu is not None:
            images = images.cuda(args.gpu, non_blocking=True)
        if torch.cuda.is_available():
            target = target.cuda(args.gpu, non_blocking=True)

        # compute output
        output = model(images)
        loss = criterion(output, target)

        # measure accuracy and record loss
        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        losses.update(loss.item(), images.size(0))
        top1.update(acc1[0], images.size(0))
        top5.update(acc5[0], images.size(0))

        # compute gradient and do SGD step
        optimizer.zero_grad()
        if args.optimizer == 'adahessian':
            loss.backward(create_graph=True)
        else:
            loss.backward()
        optimizer.step()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            progress.display(i)


def validate(pid, val_loader, model, criterion, args, validator):
    batch_time = AverageMeter('Time', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        val_loader.length,
        [batch_time, losses, top1, top5],
        prefix='pid: {} Test: '.format(pid))

    # switch to evaluate mode
    model.eval()

    with torch.no_grad():
        end = time.time()
        for i, (images, target) in enumerate(val_loader):
            if args.gpu is not None:
                images = images.cuda(args.gpu, non_blocking=True)
            target = target.cuda(args.gpu, non_blocking=True)

            # compute output
            output = model(images)
            loss = criterion(output, target)

            # measure accuracy and record loss
            acc1, acc5 = accuracy(output, target, topk=(1, 5))
            losses.update(loss.item(), images.size(0))
            top1.update(acc1[0], images.size(0))
            top5.update(acc5[0], images.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if i % args.print_freq == 0:
                progress.display(i)

        # TODO: this should also be done with the ProgressMeter
        logging.info(' * Acc@1 {top1.avg:.3f} Acc@5 {top5.avg:.3f}'
              .format(top1=top1, top5=top5))
        validator.collect_accu_result(pid, top1, top5)

    return top1.avg


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, 'model_best.pth.tar')


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        logging.info('\t'.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def adjust_learning_rate(optimizer, epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    lr = args.lr * (0.1 ** (epoch // 30))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.reshape(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size).cpu())
        return res


if __name__ == '__main__':
    torch.multiprocessing.set_start_method('spawn')
    main()
